@@ -156,10 +156,18 @@ Base.:(==)(r::Result, s::Symbol) = s == r
 ############################################################################
 # wrapper around nlopt_opt type
 
+macro const_field(expr)
+    if VERSION >= v"1.8.0-"
+        Expr(:const, esc(expr))
+    else
+        esc(expr)
+    end
+end
+
 # pass both f and o to the callback so that we can handle exceptions
-mutable struct Callback_Data
-    f::Function
-    o::Any # should be Opt, but see Julia issue #269
+mutable struct Callback_Data{F,O}
+    @const_field f::F
+    @const_field o::O # Opt
 end
 
 function Base.unsafe_convert(::Type{Ptr{Cvoid}}, c::Callback_Data)
@@ -171,13 +179,28 @@ mutable struct Opt
 
     # need to store callback data for objective and constraints in
     # Opt so that they aren't garbage-collected.  cb[1] is the objective.
-    cb::Vector{Callback_Data}
+    @const_field cb::Vector{Callback_Data}
 
     exception::Any
 
+    # These cached arrays should only be used in the callback
+    # if their pointer value and size matches the ones passed into the callback
+    x_cache::Vector{Cdouble}
+    res_cache::Vector{Cdouble}
+    grad_cache::Vector{Cdouble}
+    grad2_cache::Matrix{Cdouble}
+
     function Opt(p::Ptr{Cvoid})
         @assert p != C_NULL
-        opt = new(p, Array{Callback_Data}(undef, 1), nothing)
+        opt = new(
+            p,
+            Array{Callback_Data}(undef, 1),
+            nothing,
+            _EMPTY_VECTOR,
+            _EMPTY_VECTOR,
+            _EMPTY_VECTOR,
+            _EMPTY_MATRIX,
+        )
         finalizer(destroy, opt)
         return opt
     end
@@ -222,8 +245,8 @@ function Base.copy(opt::Opt)
     end
     new_opt = Opt(p)
     opt_callbacks = getfield(opt, :cb)
-    new_callbacks = Vector{Callback_Data}(undef, length(opt_callbacks))
-    setfield!(new_opt, :cb, new_callbacks)
+    new_callbacks = getfield(new_opt, :cb)
+    resize!(new_callbacks, length(opt_callbacks))
     old_to_new_pointer_map = Dict{Ptr{Cvoid},Ptr{Cvoid}}(C_NULL => C_NULL)
     for i in 1:length(opt_callbacks)
         if isassigned(opt_callbacks, i)
@@ -453,41 +476,74 @@ srand_time() = nlopt_srand_time()
 ############################################################################
 # Objective function:
 
+const _EMPTY_VECTOR = Cdouble[]
+const _EMPTY_MATRIX = zeros(Cdouble, 0, 0)
+
+@inline function _get_empty_vector()
+    if !isempty(_EMPTY_VECTOR)
+        empty!(_EMPTY_VECTOR)  # Reset for future calls
+        error(
+            "The builtin _EMPTY_VECTOR was modified by the user. " *
+            "If the gradient vector is empty, do not modify it in a callback.",
+        )
+    end
+    return _EMPTY_VECTOR
+end
+
+@inline function _wrap_pointer(
+    o::Opt,
+    p::Ptr,
+    sz::Tuple,
+    field::Symbol,
+    isgrad::Bool,
+)
+    cache = getfield(o, field)
+    if p == pointer(cache) && size(cache) == sz
+        return cache
+    elseif isgrad && p == C_NULL
+        return length(sz) == 1 ? _get_empty_vector() : _EMPTY_MATRIX
+    else
+        cache = unsafe_wrap(Array, p, sz)
+        setfield!(o, field, cache)
+        return cache
+    end
+end
+
 function nlopt_callback_wrapper(
     n::Cuint,
     p_x::Ptr{Cdouble},
     p_grad::Ptr{Cdouble},
-    d_::Ptr{Cvoid},
+    d::Callback_Data,
 )::Cdouble
-    d = unsafe_pointer_to_objref(d_)::Callback_Data
-    x = unsafe_wrap(Array, p_x, (n,))
-    grad = unsafe_wrap(Array, p_grad, (n,))
+    o = d.o
+    x = _wrap_pointer(o, p_x, (n,), :x_cache, false)
+    grad = _wrap_pointer(o, p_grad, (n,), :grad_cache, true)
     try
-        return d.f(x, p_grad == C_NULL ? Cdouble[] : grad)
+        return d.f(x, grad)
     catch e
-        _catch_forced_stop(d.o, e)
+        _catch_forced_stop(o, e)
     end
     return NaN
 end
 
-function min_objective!(o::Opt, f::Function)
-    cb = Callback_Data(f, o)
+function min_objective!(o::Opt, f::F) where {F}
+    cb = Callback_Data{F,Opt}(f, o)
     getfield(o, :cb)[1] = cb
     c_fn = @cfunction(
         nlopt_callback_wrapper,
         Cdouble,
-        (Cuint, Ptr{Cdouble}, Ptr{Cdouble}, Ptr{Cvoid})
+        (Cuint, Ptr{Cdouble}, Ptr{Cdouble}, Ref{Callback_Data{F,Opt}})
     )
     return chk(o, nlopt_set_min_objective(o, c_fn, cb))
 end
 
-function max_objective!(o::Opt, f::Function)
-    cb = Callback_Data(f, o)
+function max_objective!(o::Opt, f::F) where {F}
+    cb = Callback_Data{F,Opt}(f, o)
     getfield(o, :cb)[1] = cb
     c_fn = @cfunction(
         nlopt_callback_wrapper,
         Cdouble,
-        (Cuint, Ptr{Cdouble}, Ptr{Cdouble}, Ptr{Cvoid})
+        (Cuint, Ptr{Cdouble}, Ptr{Cdouble}, Ref{Callback_Data{F,Opt}})
     )
     return chk(o, nlopt_set_max_objective(o, c_fn, cb))
 end
@@ -495,24 +551,24 @@ end
 ############################################################################
 # Nonlinear constraints:
 
-function inequality_constraint!(o::Opt, f::Function, tol::Real = 0.0)
-    cb = Callback_Data(f, o)
+function inequality_constraint!(o::Opt, f::F, tol::Real = 0.0) where {F}
+    cb = Callback_Data{F,Opt}(f, o)
     push!(getfield(o, :cb), cb)
     c_fn = @cfunction(
         nlopt_callback_wrapper,
         Cdouble,
-        (Cuint, Ptr{Cdouble}, Ptr{Cdouble}, Ptr{Cvoid})
+        (Cuint, Ptr{Cdouble}, Ptr{Cdouble}, Ref{Callback_Data{F,Opt}})
     )
     return chk(o, nlopt_add_inequality_constraint(o, c_fn, cb, tol))
 end
 
-function equality_constraint!(o::Opt, f::Function, tol::Real = 0.0)
-    cb = Callback_Data(f, o)
+function equality_constraint!(o::Opt, f::F, tol::Real = 0.0) where {F}
+    cb = Callback_Data{F,Opt}(f, o)
     push!(getfield(o, :cb), cb)
     c_fn = @cfunction(
         nlopt_callback_wrapper,
         Cdouble,
-        (Cuint, Ptr{Cdouble}, Ptr{Cdouble}, Ptr{Cvoid})
+        (Cuint, Ptr{Cdouble}, Ptr{Cdouble}, Ref{Callback_Data{F,Opt}})
     )
     return chk(o, nlopt_add_equality_constraint(o, c_fn, cb, tol))
 end
@@ -533,18 +589,16 @@ function nlopt_vcallback_wrapper(
     n::Cuint,
     p_x::Ptr{Cdouble},
     p_grad::Ptr{Cdouble},
-    d_::Ptr{Cvoid},
+    d::Callback_Data,
 )
-    d = unsafe_pointer_to_objref(d_)::Callback_Data
-    res = unsafe_wrap(Array, p_res, (m,))
-    x = unsafe_wrap(Array, p_x, (n,))
-    grad =
-        p_grad == C_NULL ? zeros(Cdouble, 0, 0) :
-        unsafe_wrap(Array, p_grad, (n, m))
+    o = d.o
+    res = _wrap_pointer(o, p_res, (m,), :res_cache, false)
+    x = _wrap_pointer(o, p_x, (n,), :x_cache, false)
+    grad = _wrap_pointer(o, p_grad, (n, m), :grad2_cache, true)
     try
         d.f(res, x, grad)
     catch e
-        _catch_forced_stop(d.o, e)
+        _catch_forced_stop(o, e)
     end
     return
 end
@@ -561,51 +615,56 @@ function _catch_forced_stop(o::Opt, e)
     return
 end
 
-function inequality_constraint!(o::Opt, f::Function, tol::Vector{Cdouble})
-    cb = Callback_Data(f, o)
+function inequality_constraint!(o::Opt, f::F, tol::Vector{Cdouble}) where {F}
+    cb = Callback_Data{F,Opt}(f, o)
     push!(getfield(o, :cb), cb)
     c_fn = @cfunction(
         nlopt_vcallback_wrapper,
         Cvoid,
-        (Cuint, Ptr{Cdouble}, Cuint, Ptr{Cdouble}, Ptr{Cdouble}, Ptr{Cvoid}),
+        (
+            Cuint,
+            Ptr{Cdouble},
+            Cuint,
+            Ptr{Cdouble},
+            Ptr{Cdouble},
+            Ref{Callback_Data{F,Opt}},
+        ),
     )
     ret = nlopt_add_inequality_mconstraint(o, length(tol), c_fn, cb, tol)
     return chk(o, ret)
 end
 
-function inequality_constraint!(
-    o::Opt,
-    f::Function,
-    tol::AbstractVector{<:Real},
-)
+function inequality_constraint!(o::Opt, f, tol::AbstractVector{<:Real})
     return inequality_constraint!(o, f, Array{Float64}(tol))
 end
 
-function inequality_constraint!(
-    o::Opt,
-    m::Integer,
-    f::Function,
-    tol::Real = 0.0,
-)
+function inequality_constraint!(o::Opt, m::Integer, f, tol::Real = 0.0)
     return inequality_constraint!(o, f, fill(Cdouble(tol), m))
 end
 
-function equality_constraint!(o::Opt, f::Function, tol::Vector{Cdouble})
-    cb = Callback_Data(f, o)
+function equality_constraint!(o::Opt, f::F, tol::Vector{Cdouble}) where {F}
+    cb = Callback_Data{F,Opt}(f, o)
     push!(getfield(o, :cb), cb)
     c_fn = @cfunction(
         nlopt_vcallback_wrapper,
         Cvoid,
-        (Cuint, Ptr{Cdouble}, Cuint, Ptr{Cdouble}, Ptr{Cdouble}, Ptr{Cvoid}),
+        (
+            Cuint,
+            Ptr{Cdouble},
+            Cuint,
+            Ptr{Cdouble},
+            Ptr{Cdouble},
+            Ref{Callback_Data{F,Opt}},
+        ),
     )
     return chk(o, nlopt_add_equality_mconstraint(o, length(tol), c_fn, cb, tol))
 end
 
-function equality_constraint!(o::Opt, f::Function, tol::AbstractVector{<:Real})
+function equality_constraint!(o::Opt, f, tol::AbstractVector{<:Real})
     return equality_constraint!(o, f, Array{Float64}(tol))
 end
 
-function equality_constraint!(o::Opt, m::Integer, f::Function, tol::Real = 0.0)
+function equality_constraint!(o::Opt, m::Integer, f, tol::Real = 0.0)
     return equality_constraint!(o, f, fill(Cdouble(tol), m))
 end
 
